//{{NO_DEPENDENCIES}}
// Microsoft Visual C++ generated include file.
// Used by resource.rc
//
#define IDR_COPY_PS                     101
#define IDR_FULLSCREEN_VS               102
#define IDR_IMGUI_PS_3_0                103
#define IDR_IMGUI_PS_4_0                104
#define IDR_IMGUI_PS_4_0_HDR            105
#define IDR_IMGUI_PS_SPIRV              106
#define IDR_IMGUI_PS_SPIRV_HDR          107
#define IDR_IMGUI_VS_3_0                108
#define IDR_IMGUI_VS_4_0                109
#define IDR_IMGUI_VS_SPIRV              110
#define IDR_MIPMAP_CS                   111
#define IDB_MAIN_ICON                   112
#define IDR_LICENSE_GL3W                701
#define IDR_LICENSE_IMGUI               702
#define IDR_LICENSE_MINHOOK             703
#define IDR_LICENSE_OPENVR              704
#define IDR_LICENSE_OPENXR              705
#define IDR_LICENSE_RESHADE             706
#define IDR_LICENSE_SPIRV               707
#define IDR_LICENSE_STB                 708
#define IDR_LICENSE_UTFCPP              709
#define IDR_LICENSE_VMA                 710
#define IDR_LICENSE_VULKAN              711

// Next default values for new objects
// 
#ifdef APSTUDIO_INVOKED
#ifndef APSTUDIO_READONLY_SYMBOLS
#define _APS_NEXT_RESOURCE_VALUE        113
#define _APS_NEXT_COMMAND_VALUE         40001
#define _APS_NEXT_CONTROL_VALUE         1001
#define _APS_NEXT_SYMED_VALUE           113
#endif
#endif
