#include "forkawesome.h"

static const char FONT_ICON_BUFFER_NAME_FK[3110+1] =
    "7])#######l4F<6'/###[),##+Sl##Q6>##w#S+HcY..Xm1NP&&A'o/dJ[^Ii4j)%afG<-dwhP/-2JuBXhc/(,xkA##B^01:_[FH<:p3#L,>>#:qEn/Bd@UCAuOi$hhG<-87T;-QF%U."
    "4tw<6rhG<-]vhP/Q-0%JU)*Im0YK_#<umo%3OV=BaFA>:7KwA--4S>-Mr*J-2%KgL`Kss'k,e'&(_.R3dFrZu2/FI%[CP)M*f(HMP;7A-A_@U.Blu##[brX-rLmHHsn&W74xD_&4oP]O"
    ":*L/L7q+N'w[a`WCT8U)$bj-$D)+Z$CFt-$N;>uu`0YR*OWe;%s9]jL3MG8.'5>##SFp&#3;G##WPj&#%r0hLgVV8&jIdK'<;/q-X<haD>,-9dS=N'*ABppd#),##i@H0'L0:/DNAhJ)"
    "Fs.O=/MbA#99uD+rSXL#=1GcMgJG&#i8Lw-Sr[fLrrH##.GY##2Sl##6`($#ePgc.>xL$#1ZqV%p[%/1,60MTqWLA+%O5SeGDfY,os.Gik/f`E-PQ/1<l,F.3?iu5:Bqc)Hu5X:8.]f1"
    "Z&x+2s6+<$J_nt[G+WR6^']fLVvIo[5UV=->?kR/9hu>#fl?Q#(./J36<Xq.]3@#6`v-MB$'pIU$p:u>qiYLMOP?uuYg8e$UNc##K(m<-?qD#MbPMG)@f)T/^b%T%S2`^#J:8Y.pV*i("
    "T=)?#h-[guT#9iu&](?#A]wG;[Dm]uB*07QK(]qFV=fV$H[`V$#kUK#$8^fLmw@8%P92^uJ98=/+@u2OHI.o`5BOojOps+/QlRfL*xFt$=bx5#G%KgLmb]s$wkH>#iE(E#VA@r%Mep$-"
    "T1G$,G2J1,mQOhuvne.Mv?75/Huiw'6`?$=VC]&,EH*7M:9s%,:7QVQNcb-?G<fT$_LT2#kiKS.M7%s$MhGgLN7LkL4%du%LE,-*Hx:?#TTPe?u@BquCEB_-+Md'&4:(B#KjJgM18YCj"
    "p9sQNjw/SS>lg<M:D$##&&>uuWN3L#Y@%%#+.>15XAic)rkh8.mNv)4')'J3Ne_F*-vY8/NmVE-*tbU%.qFX$]f.'5/^@+0j.8$$H33-,+VxA&W@1,'<Zd1B,o$>#(kZoAKkP9MTC(T7"
    "t#;(,b_mL,v?qr&4uG##%J/GV;leJj:SU*%S6[0#9%T*#?3B:%isAv-&O]s$^jtJ)iMTG.KG>c4[lZ#6*@G.)Kfhd3WH5<.*M4gL;ngo0TAOZ6W.PV-RVx_4QmuX$KOj`ub>(6RTcZUL"
    "bR39/oBfU+-p9x9SP>bnQ5&X%pbO-)&-g[us_.u-=^7>N8$EG'i_7MTv+qM#ReBtLg@RxktpqbrUv?rC*+8C8i&X1)e61nLx-$O*%@h-Ml=DX-xv@=(`()?#B8WeFeN4kL4Sq'#'&>uu"
    "D@]p$?HY##Pe[%#jMOV-H,Jd)K?_`3qh)?#0SEF3oc@A4voXV-HNpdmrZ*G4S)ZA#V8HC@BO?h#l7=9%8><gL)7hY,Wt/GV@wg1B5QwDXMx6f.KL$Toj&LW.^aDGVYmkv-x'5gLH.Tou"
    "BXqR#1Rh=4;r3GVUm^8oK,>>#XIVS%Ae8e$rLb&#MYWI))CuM(jAqB#]p*P(uAOZ6m('J3k#fF4Rq@.*G7K,3L1O058b4-5mS4'5841'5%J/GV/1:B#'at%$q^DIMiYV6MdS6##,>3/L"
    "dbb(NRdimM-a4o7qPa>$cMDX:c)L'5&8YuuCmxE%jxIfLNdP&#&sgo.5>PjLcgp>,&'ihLoGb%$kLu/MdB]s$8uaX(g?T:%-<3jLV*<9/r4)mL`ZH##gFYF'rh%T%kE,7&#-xh&`]3I)"
    "bRX-)Vi08.bC/#.&_H#6*,`#$b9bS9#Yhb'0R:-MA<6GMXfU9%NB5N'q(73'4(NSt?fxD,os3;7]OQ--#E^>e&@,gL7<wu#vAYc2iH%##*c?X-GOFb3J?gf1sLs>R=g*P(Bis9))%x[-"
    ",CS@#_t1kbE'7eM=WqP8[`_s-D7H>#Ot]=u'bK<L(ieRnO_K$52g@+0?PS05bnTs/f^eNOUB;=-+gJc$+^tpAu_L%,w$[gu5[-['#**&+-?0aYWLh=#%H3'50S6##/4K-)Xf?%b5W^/)"
    "m`):)t1@k==?#]u3i8-#$),##7iKWJ2:>d33_+/(R4U8.b(<9/%&AA4LU%],sq-x64=S_#vBo8%FLHm8=N#B=[5ZA#83E.3C4Z##PIpUm6MU7eK2[S%+.`0(k3VNI%(Y4=%<D[D`m6-O"
    "995O0M1r7/i;]iL6wp9RVn]7f`0pUdPWwJJ]'+'+lbZZ#xMUqLpX[&MmpY<8q)sp/%A@`a9+`w,ICg@JI15##/qiw$)8[0#g,_'#8%Fb3co.W]0CRm/d(4I)kHL,3bG2:%@Cw8.?]J,3"
    "JxQ&5hVN<qPc;^uL=YD%XXKb@@hsPM2.VfLEF8GV<cUr#n?5C/ta]e$.=HVuAhL#$]1UW.s]'##MEe*.x<trL5Yj$#:oA*#0M#lLf]$Q8=QF$Kbc@A4.*:%Qw#pr6D?B:%c8hh$MjrD,"
    "J]B.*L2fcr=a@+0s#YCjYtlS7m.%A#8.-xOjeAE#9,<B16CW?Q94s^uSwu`8doNVM:%fV@N2Qx?RPno@:Z>]uY`n9.0O7DEhTr<M2hu`1@l3,4/>2a52Ca79XIMbevebt2o7^Bf,jl2r"
    "(2M'#J.#I9mU7C#mc4c4br9s-;va.3D<D$%DB&0ufu'L#>Y[OVk1k3Gpv,W.O-F%beUU@-$eQX.4^(##2tff.P=#+#g_</7-@XA#R4r?#fgGj'x%MT/ObE]-QtfM'-Wgf1Q<258MnP,*"
    "?ulW-tYqw0`EmS/si+C#[<;S&mInS%FAov#Fu[E+*L'O'GWuN'+)U40`a5k'sA;)*RIRF%Tw/Z--1=e?5;1X:;vPk&CdNjL*(KJ1/,TV-G(^S*v14gL#8,,M*YPgLaII@b*]IK37Ah1T"
    "GkI9(&;cY#,VC;$0o$s$9bJM'rCdS%8I<5&<bsl&@$TM'dMd<Bd*fe#ha8jBgWiTCdMBnDo1/ZGrDFVCvxD#Go&+nDpuX>B1#?LFdo>kEBbT%JpO]:Co4;hFwf1eG0Y5W-lKx?0aeKh5"
    "WTw0FZd^$&p)fUC2ve:C3n8AtjXI)g";

