EXPORTS
	; d2d1.dll (Uses ordinals in d2d1.lib for all exports)
	D2D1CreateDevice @7 PRIVATE
	D2D1CreateFactory @1 PRIVATE

	; d3d9.dll (Uses ordinals in d3d9.lib for 'Direct3DCreate9On12' and 'Direct3DCreate9On12Ex')
	Direct3DCreate9 PRIVATE
	Direct3DCreate9Ex PRIVATE
	Direct3DCreate9On12 @20 PRIVATE
	Direct3DCreate9On12Ex @21 PRIVATE

	D3DPERF_BeginEvent PRIVATE
	D3DPERF_EndEvent PRIVATE
	D3DPERF_GetStatus PRIVATE
	D3DPERF_QueryRepeatFrame PRIVATE
	D3DPERF_SetMarker PRIVATE
	D3DPERF_SetOptions PRIVATE
	D3DPERF_SetRegion PRIVATE

	Direct3D9ForceHybridEnumeration @16 NONAME PRIVATE
	Direct3D9SetMaximizedWindowedModeShim @17 NONAME PRIVATE
	Direct3D9SetSwapEffectUpgradeShim @18 NONAME PRIVATE
	Direct3D9Force9on12 @19 NONAME PRIVATE
	Direct3D9SetMaximizedWindowHwndOverride @22 NONAME PRIVATE
	Direct3D9SetVendorIDLieFor9on12 @23 NONAME PRIVATE
	Direct3D9EnableMaximizedWindowedModeShim PRIVATE

	; d3d10.dll
	D3D10CreateDevice @4 PRIVATE
	D3D10CreateDeviceAndSwapChain @5 PRIVATE

	; d3d10_1.dll
	D3D10CreateDevice1 PRIVATE
	D3D10CreateDeviceAndSwapChain1 PRIVATE

	; d3d11.dll
	D3D11CreateDevice PRIVATE
	D3D11CreateDeviceAndSwapChain PRIVATE
	D3D11On12CreateDevice PRIVATE

	; d3d12.dll (Uses ordinals in d3d12.lib for 'D3D12CreateDevice' and 'D3D12GetDebugInterface')
	D3D12CreateDevice @101 PRIVATE
	D3D12GetDebugInterface @102 PRIVATE
	D3D12CreateRootSignatureDeserializer @107 PRIVATE
	D3D12CreateVersionedRootSignatureDeserializer @108 PRIVATE
	D3D12EnableExperimentalFeatures @110 PRIVATE
	D3D12GetInterface @111 PRIVATE
	D3D12PIXEventsReplaceBlock PRIVATE
	D3D12PIXGetThreadInfo PRIVATE
	D3D12PIXNotifyWakeFromFenceSignal PRIVATE
	D3D12PIXReportCounter PRIVATE
	D3D12SerializeRootSignature PRIVATE
	D3D12SerializeVersionedRootSignature PRIVATE

	; dxgi.dll
	CreateDXGIFactory PRIVATE
	CreateDXGIFactory1 PRIVATE
	CreateDXGIFactory2 PRIVATE

	DXGID3D10CreateDevice PRIVATE
	DXGID3D10CreateLayeredDevice PRIVATE
	DXGID3D10ETWRundown PRIVATE
	DXGID3D10GetLayeredDeviceSize PRIVATE
	DXGID3D10RegisterLayers PRIVATE

	CompatValue PRIVATE
	CompatString PRIVATE
	DXGIDumpJournal PRIVATE
	DXGIDeclareAdapterRemovalSupport PRIVATE
	DXGIGetDebugInterface1 PRIVATE
	DXGIReportAdapterConfiguration PRIVATE

	; opengl32.dll
	glAccum PRIVATE
	glAlphaFunc PRIVATE
	glAreTexturesResident PRIVATE
	glArrayElement PRIVATE
	glBegin PRIVATE
	glBindTexture PRIVATE
	glBitmap PRIVATE
	glBlendFunc PRIVATE
	glCallList PRIVATE
	glCallLists PRIVATE
	glClear PRIVATE
	glClearAccum PRIVATE
	glClearColor PRIVATE
	glClearDepth PRIVATE
	glClearIndex PRIVATE
	glClearStencil PRIVATE
	glClipPlane PRIVATE
	glColor3b PRIVATE
	glColor3bv PRIVATE
	glColor3d PRIVATE
	glColor3dv PRIVATE
	glColor3f PRIVATE
	glColor3fv PRIVATE
	glColor3i PRIVATE
	glColor3iv PRIVATE
	glColor3s PRIVATE
	glColor3sv PRIVATE
	glColor3ub PRIVATE
	glColor3ubv PRIVATE
	glColor3ui PRIVATE
	glColor3uiv PRIVATE
	glColor3us PRIVATE
	glColor3usv PRIVATE
	glColor4b PRIVATE
	glColor4bv PRIVATE
	glColor4d PRIVATE
	glColor4dv PRIVATE
	glColor4f PRIVATE
	glColor4fv PRIVATE
	glColor4i PRIVATE
	glColor4iv PRIVATE
	glColor4s PRIVATE
	glColor4sv PRIVATE
	glColor4ub PRIVATE
	glColor4ubv PRIVATE
	glColor4ui PRIVATE
	glColor4uiv PRIVATE
	glColor4us PRIVATE
	glColor4usv PRIVATE
	glColorMask PRIVATE
	glColorMaterial PRIVATE
	glColorPointer PRIVATE
	glCopyPixels PRIVATE
	glCopyTexImage1D PRIVATE
	glCopyTexImage2D PRIVATE
	glCopyTexSubImage1D PRIVATE
	glCopyTexSubImage2D PRIVATE
	glCullFace PRIVATE
	glDeleteLists PRIVATE
	glDeleteTextures PRIVATE
	glDepthFunc PRIVATE
	glDepthMask PRIVATE
	glDepthRange PRIVATE
	glDisable PRIVATE
	glDisableClientState PRIVATE
	glDrawArrays PRIVATE
	glDrawBuffer PRIVATE
	glDrawElements PRIVATE
	glDrawPixels PRIVATE
	glEdgeFlag PRIVATE
	glEdgeFlagPointer PRIVATE
	glEdgeFlagv PRIVATE
	glEnable PRIVATE
	glEnableClientState PRIVATE
	glEnd PRIVATE
	glEndList PRIVATE
	glEvalCoord1d PRIVATE
	glEvalCoord1dv PRIVATE
	glEvalCoord1f PRIVATE
	glEvalCoord1fv PRIVATE
	glEvalCoord2d PRIVATE
	glEvalCoord2dv PRIVATE
	glEvalCoord2f PRIVATE
	glEvalCoord2fv PRIVATE
	glEvalMesh1 PRIVATE
	glEvalMesh2 PRIVATE
	glEvalPoint1 PRIVATE
	glEvalPoint2 PRIVATE
	glFeedbackBuffer PRIVATE
	glFinish PRIVATE
	glFlush PRIVATE
	glFogf PRIVATE
	glFogfv PRIVATE
	glFogi PRIVATE
	glFogiv PRIVATE
	glFrontFace PRIVATE
	glFrustum PRIVATE
	glGenLists PRIVATE
	glGenTextures PRIVATE
	glGetBooleanv PRIVATE
	glGetClipPlane PRIVATE
	glGetDoublev PRIVATE
	glGetError PRIVATE
	glGetFloatv PRIVATE
	glGetIntegerv PRIVATE
	glGetLightfv PRIVATE
	glGetLightiv PRIVATE
	glGetMapdv PRIVATE
	glGetMapfv PRIVATE
	glGetMapiv PRIVATE
	glGetMaterialfv PRIVATE
	glGetMaterialiv PRIVATE
	glGetPixelMapfv PRIVATE
	glGetPixelMapuiv PRIVATE
	glGetPixelMapusv PRIVATE
	glGetPointerv PRIVATE
	glGetPolygonStipple PRIVATE
	glGetString PRIVATE
	glGetTexEnvfv PRIVATE
	glGetTexEnviv PRIVATE
	glGetTexGendv PRIVATE
	glGetTexGenfv PRIVATE
	glGetTexGeniv PRIVATE
	glGetTexImage PRIVATE
	glGetTexLevelParameterfv PRIVATE
	glGetTexLevelParameteriv PRIVATE
	glGetTexParameterfv PRIVATE
	glGetTexParameteriv PRIVATE
	glHint PRIVATE
	glIndexMask PRIVATE
	glIndexPointer PRIVATE
	glIndexd PRIVATE
	glIndexdv PRIVATE
	glIndexf PRIVATE
	glIndexfv PRIVATE
	glIndexi PRIVATE
	glIndexiv PRIVATE
	glIndexs PRIVATE
	glIndexsv PRIVATE
	glIndexub PRIVATE
	glIndexubv PRIVATE
	glInitNames PRIVATE
	glInterleavedArrays PRIVATE
	glIsEnabled PRIVATE
	glIsList PRIVATE
	glIsTexture PRIVATE
	glLightModelf PRIVATE
	glLightModelfv PRIVATE
	glLightModeli PRIVATE
	glLightModeliv PRIVATE
	glLightf PRIVATE
	glLightfv PRIVATE
	glLighti PRIVATE
	glLightiv PRIVATE
	glLineStipple PRIVATE
	glLineWidth PRIVATE
	glListBase PRIVATE
	glLoadIdentity PRIVATE
	glLoadMatrixd PRIVATE
	glLoadMatrixf PRIVATE
	glLoadName PRIVATE
	glLogicOp PRIVATE
	glMap1d PRIVATE
	glMap1f PRIVATE
	glMap2d PRIVATE
	glMap2f PRIVATE
	glMapGrid1d PRIVATE
	glMapGrid1f PRIVATE
	glMapGrid2d PRIVATE
	glMapGrid2f PRIVATE
	glMaterialf PRIVATE
	glMaterialfv PRIVATE
	glMateriali PRIVATE
	glMaterialiv PRIVATE
	glMatrixMode PRIVATE
	glMultMatrixd PRIVATE
	glMultMatrixf PRIVATE
	glNewList PRIVATE
	glNormal3b PRIVATE
	glNormal3bv PRIVATE
	glNormal3d PRIVATE
	glNormal3dv PRIVATE
	glNormal3f PRIVATE
	glNormal3fv PRIVATE
	glNormal3i PRIVATE
	glNormal3iv PRIVATE
	glNormal3s PRIVATE
	glNormal3sv PRIVATE
	glNormalPointer PRIVATE
	glOrtho PRIVATE
	glPassThrough PRIVATE
	glPixelMapfv PRIVATE
	glPixelMapuiv PRIVATE
	glPixelMapusv PRIVATE
	glPixelStoref PRIVATE
	glPixelStorei PRIVATE
	glPixelTransferf PRIVATE
	glPixelTransferi PRIVATE
	glPixelZoom PRIVATE
	glPointSize PRIVATE
	glPolygonMode PRIVATE
	glPolygonOffset PRIVATE
	glPolygonStipple PRIVATE
	glPopAttrib PRIVATE
	glPopClientAttrib PRIVATE
	glPopMatrix PRIVATE
	glPopName PRIVATE
	glPrioritizeTextures PRIVATE
	glPushAttrib PRIVATE
	glPushClientAttrib PRIVATE
	glPushMatrix PRIVATE
	glPushName PRIVATE
	glRasterPos2d PRIVATE
	glRasterPos2dv PRIVATE
	glRasterPos2f PRIVATE
	glRasterPos2fv PRIVATE
	glRasterPos2i PRIVATE
	glRasterPos2iv PRIVATE
	glRasterPos2s PRIVATE
	glRasterPos2sv PRIVATE
	glRasterPos3d PRIVATE
	glRasterPos3dv PRIVATE
	glRasterPos3f PRIVATE
	glRasterPos3fv PRIVATE
	glRasterPos3i PRIVATE
	glRasterPos3iv PRIVATE
	glRasterPos3s PRIVATE
	glRasterPos3sv PRIVATE
	glRasterPos4d PRIVATE
	glRasterPos4dv PRIVATE
	glRasterPos4f PRIVATE
	glRasterPos4fv PRIVATE
	glRasterPos4i PRIVATE
	glRasterPos4iv PRIVATE
	glRasterPos4s PRIVATE
	glRasterPos4sv PRIVATE
	glReadBuffer PRIVATE
	glReadPixels PRIVATE
	glRectd PRIVATE
	glRectdv PRIVATE
	glRectf PRIVATE
	glRectfv PRIVATE
	glRecti PRIVATE
	glRectiv PRIVATE
	glRects PRIVATE
	glRectsv PRIVATE
	glRenderMode PRIVATE
	glRotated PRIVATE
	glRotatef PRIVATE
	glScaled PRIVATE
	glScalef PRIVATE
	glScissor PRIVATE
	glSelectBuffer PRIVATE
	glShadeModel PRIVATE
	glStencilFunc PRIVATE
	glStencilMask PRIVATE
	glStencilOp PRIVATE
	glTexCoord1d PRIVATE
	glTexCoord1dv PRIVATE
	glTexCoord1f PRIVATE
	glTexCoord1fv PRIVATE
	glTexCoord1i PRIVATE
	glTexCoord1iv PRIVATE
	glTexCoord1s PRIVATE
	glTexCoord1sv PRIVATE
	glTexCoord2d PRIVATE
	glTexCoord2dv PRIVATE
	glTexCoord2f PRIVATE
	glTexCoord2fv PRIVATE
	glTexCoord2i PRIVATE
	glTexCoord2iv PRIVATE
	glTexCoord2s PRIVATE
	glTexCoord2sv PRIVATE
	glTexCoord3d PRIVATE
	glTexCoord3dv PRIVATE
	glTexCoord3f PRIVATE
	glTexCoord3fv PRIVATE
	glTexCoord3i PRIVATE
	glTexCoord3iv PRIVATE
	glTexCoord3s PRIVATE
	glTexCoord3sv PRIVATE
	glTexCoord4d PRIVATE
	glTexCoord4dv PRIVATE
	glTexCoord4f PRIVATE
	glTexCoord4fv PRIVATE
	glTexCoord4i PRIVATE
	glTexCoord4iv PRIVATE
	glTexCoord4s PRIVATE
	glTexCoord4sv PRIVATE
	glTexCoordPointer PRIVATE
	glTexEnvf PRIVATE
	glTexEnvfv PRIVATE
	glTexEnvi PRIVATE
	glTexEnviv PRIVATE
	glTexGend PRIVATE
	glTexGendv PRIVATE
	glTexGenf PRIVATE
	glTexGenfv PRIVATE
	glTexGeni PRIVATE
	glTexGeniv PRIVATE
	glTexImage1D PRIVATE
	glTexImage2D PRIVATE
	glTexParameterf PRIVATE
	glTexParameterfv PRIVATE
	glTexParameteri PRIVATE
	glTexParameteriv PRIVATE
	glTexSubImage1D PRIVATE
	glTexSubImage2D PRIVATE
	glTranslated PRIVATE
	glTranslatef PRIVATE
	glVertex2d PRIVATE
	glVertex2dv PRIVATE
	glVertex2f PRIVATE
	glVertex2fv PRIVATE
	glVertex2i PRIVATE
	glVertex2iv PRIVATE
	glVertex2s PRIVATE
	glVertex2sv PRIVATE
	glVertex3d PRIVATE
	glVertex3dv PRIVATE
	glVertex3f PRIVATE
	glVertex3fv PRIVATE
	glVertex3i PRIVATE
	glVertex3iv PRIVATE
	glVertex3s PRIVATE
	glVertex3sv PRIVATE
	glVertex4d PRIVATE
	glVertex4dv PRIVATE
	glVertex4f PRIVATE
	glVertex4fv PRIVATE
	glVertex4i PRIVATE
	glVertex4iv PRIVATE
	glVertex4s PRIVATE
	glVertex4sv PRIVATE
	glVertexPointer PRIVATE
	glViewport PRIVATE
	
	wglChoosePixelFormat PRIVATE
	wglCopyContext PRIVATE
	wglCreateContext PRIVATE
	wglCreateLayerContext PRIVATE
	wglDeleteContext PRIVATE
	wglDescribeLayerPlane PRIVATE
	wglDescribePixelFormat PRIVATE
	wglGetCurrentContext PRIVATE
	wglGetCurrentDC PRIVATE
	wglGetLayerPaletteEntries PRIVATE
	wglGetPixelFormat PRIVATE
	wglGetProcAddress PRIVATE
	wglMakeCurrent PRIVATE
	wglRealizeLayerPalette PRIVATE
	wglSetLayerPaletteEntries PRIVATE
	wglSetPixelFormat PRIVATE
	wglShareLists PRIVATE
	wglSwapBuffers PRIVATE
	wglSwapLayerBuffers PRIVATE
	wglSwapMultipleBuffers PRIVATE
	wglUseFontBitmapsA PRIVATE
	wglUseFontBitmapsW PRIVATE
	wglUseFontOutlinesA PRIVATE
	wglUseFontOutlinesW PRIVATE

	; openxr_loader.dll
	xrNegotiateLoaderApiLayerInterface

	; vrclient[_x64].dll
	VRClientCoreFactory PRIVATE

	; vulkan-1.dll
	vkNegotiateLoaderLayerInterfaceVersion
	vkCreateDevice PRIVATE
	vkCreateInstance PRIVATE
	vkGetDeviceProcAddr PRIVATE
	vkGetInstanceProcAddr PRIVATE
	vkGetPhysicalDeviceToolPropertiesEXT PRIVATE

	; user32.dll
	RegisterClassA = HookRegisterClassA PRIVATE
	RegisterClassW = HookRegisterClassW PRIVATE
	RegisterClassExA = HookRegisterClassExA PRIVATE
	RegisterClassExW = HookRegisterClassExW PRIVATE
	RegisterRawInputDevices = HookRegisterRawInputDevices PRIVATE
	GetMessageA = HookGetMessageA PRIVATE
	GetMessageW = HookGetMessageW PRIVATE
	PeekMessageA = HookPeekMessageA PRIVATE
	PeekMessageW = HookPeekMessageW PRIVATE
	PostMessageA = HookPostMessageA PRIVATE
	PostMessageW = HookPostMessageW PRIVATE
	ClipCursor = HookClipCursor PRIVATE
	GetCursorPos = HookGetCursorPosition PRIVATE
	SetCursorPos = HookSetCursorPosition PRIVATE

	; ws2_32.dll (Uses ordinals in ws2_32.lib for most exports)
	send = HookSend PRIVATE
	sendto = HookSendTo PRIVATE
	recv = HookRecv PRIVATE
	recvfrom = HookRecvFrom PRIVATE
	WSASend = HookWSASend @96 PRIVATE
	WSASendTo = HookWSASendTo @99 PRIVATE
	WSARecv = HookWSARecv @91 PRIVATE
	WSARecvFrom = HookWSARecvFrom @93 PRIVATE

	; dinput.dll
	DirectInputCreateA PRIVATE
	DirectInputCreateW PRIVATE
	DirectInputCreateEx PRIVATE

	; dinput8.dll
	DirectInput8Create PRIVATE
	GetdfDIJoystick PRIVATE
