// The MIT License (MIT)

// Copyright (c) 2013-2020 Rapptz, ThePhD and contributors

// Permission is hereby granted, free of charge, to any person obtaining a copy of
// this software and associated documentation files (the "Software"), to deal in
// the Software without restriction, including without limitation the rights to
// use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
// the Software, and to permit persons to whom the Software is furnished to do so,
// subject to the following conditions:

// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.

// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
// FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
// COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
// IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
// CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

// This file was generated with a script.
// Generated 2024-09-03 03:46:49.901324 UTC
// This header was generated with sol v3.3.1 (revision 2b0d2fe8)
// https://github.com/ThePhD/sol2

#ifndef SOL_SINGLE_SOL_CONFIG_HPP
#define SOL_SINGLE_SOL_CONFIG_HPP

// beginning of sol/config.hpp

/* Base, empty configuration file!

     To override, place a file in your include paths of the form:

. (your include path here)
| sol (directory, or equivalent)
  | config.hpp (your config.hpp file)

     So that when sol2 includes the file

#include <sol/config.hpp>

     it gives you the configuration values you desire. Configuration values can be
seen in the safety.rst of the doc/src, or at
https://sol2.readthedocs.io/en/latest/safety.html ! You can also pass them through
the build system, or the command line options of your compiler.

*/

#define SOL_NO_EXCEPTIONS 1

// end of sol/config.hpp

#endif // SOL_SINGLE_SOL_CONFIG_HPP
