<Page x:Class="ReShade.Setup.Pages.SelectOperationPage"
      xmlns="http://schemas.microsoft.com/winfx/2006/xaml/presentation"
      xmlns:x="http://schemas.microsoft.com/winfx/2006/xaml"
      xmlns:mc="http://schemas.openxmlformats.org/markup-compatibility/2006" 
      xmlns:d="http://schemas.microsoft.com/expression/blend/2008"
      mc:Ignorable="d" 
      d:DesignWidth="510" d:DesignHeight="638"
      Width="Auto" Height="Auto"
      FocusManager.FocusedElement="{Binding ElementName=UpdateButton}">
  <DockPanel>
    <Label x:Uid="PageTitle" DockPanel.Dock="Top" Margin="10" Padding="0" VerticalAlignment="Top" Foreground="#FF003399" FontSize="16" Content="Select the operation to perform:" />
    <TextBlock x:Uid="PageDescription" DockPanel.Dock="Top" Margin="10,5" TextWrapping="Wrap">
      An existing ReShade installation was found for the target game. Choose how to proceed with this installation.
    </TextBlock>
    <StackPanel Orientation="Vertical" HorizontalAlignment="Center" VerticalAlignment="Center">
      <RadioButton x:Name="UpdateButton" x:FieldModifier="private" Margin="0,10" IsChecked="True">
        <TextBlock x:Uid="UpdateButton" Width="290" TextWrapping="Wrap">
          <Label Content="Update _ReShade only" Padding="0" Foreground="#FF003399" FontSize="16" />
          <LineBreak />
          <LineBreak />
          Update ReShade to this version, but keep existing preset and effect files unchanged.
        </TextBlock>
      </RadioButton>
      <RadioButton x:Name="ModifyButton" x:FieldModifier="private" Margin="0,10">
        <TextBlock x:Uid="ModifyButton" Width="290" TextWrapping="Wrap">
          <Label Content="Update ReShade and _effects" Padding="0" Foreground="#FF003399" FontSize="16" />
          <LineBreak />
          <LineBreak />
          Update ReShade to this version and also download effect files again.
        </TextBlock>
      </RadioButton>
      <RadioButton x:Name="UninstallButton" x:FieldModifier="private" Margin="0,10" VerticalContentAlignment="Top">
        <TextBlock x:Uid="UninstallButton" Width="290" TextWrapping="Wrap">
          <Label Content="_Uninstall ReShade and effects" Padding="0" Foreground="#FF003399" FontSize="16" />
          <LineBreak />
          <LineBreak />
          Remove ReShade, related configuration, logs and all installed effect files from the target game.
        </TextBlock>
      </RadioButton>
    </StackPanel>
  </DockPanel>
</Page>
